% Main script.
% Use 7, 8, 9, 4, 5, 6 for movement, m for map on/off, r for refresh map,
% s for observer position input by mouse, q for quit.

%close all
clear

global g_zero g_angle g_quit g_is_map_fig g_event

g_zero = (100*eps);

% Grid cell size.
gcs = [64; 64];

% Grid sizes.
ngx = 10;
ngy = 10;
ng = [ngx; ngy];

% Movement params.
mov.turn = 7;
mov.strafe = 20;
mov.forward = 20;
mov.backward = -20;

% Texture height.
tex_h = 128;

% Texture.
%texture = rand( gcs(1), tex_h, 3 );
img = imread( 'texture3.jpg' );
texture(1,:,:,:) = double(img) / 255.0;
img = imread( 'texture1.jpg' );
texture(2,:,:,:) = double(img) / 255.0;
img = imread( 'texture2.jpg' );
texture(3,:,:,:) = double(img) / 255.0;
img = imread( 'texture4.jpg' );
texture(4,:,:,:) = double(img) / 255.0;
img = imread( 'texture5.jpg' );
texture(5,:,:,:) = double(img) / 255.0;
%texture(65:128,:,:) = texture;

% Observer point.
op = [100; 100];

% View direction angle.
oa = 45;

% Field of view semiangle.
fov.a = 30;

% Field of view size.
fov.s = [200; 100];

% Field of view screen distance.
fov.sd = 0.5 * fov.s(1) / tan( pi / 180 * fov.a );

fov

% Compute ray line angles.
aux = fov.s(1) / 2;
for i = 1:fov.s(1)
  g_angle(i) = -180 / pi * atan( (i - aux) / fov.sd );
end

% Create scale map.
smap.lmin = 1;
smap.lmax = 400;
smap.map = create_scale_map_range( tex_h, smap.lmin, smap.lmax );

% Observer point in grid coordinates.
opg = floor( op ./ gcs ) + 1;

% Local grid origin.
%lgo = opg .* gcs

% Local observer point.
%lop = op - lgo

% Grid.
grid =[1,1,1,1,1,1,0,0,1,1;
  1,0,0,0,0,3,0,0,3,1;
  1,0,0,0,0,0,0,0,0,1;
  1,0,2,0,0,0,0,0,0,1;
  1,0,2,0,0,0,4,4,0,1;
  1,0,2,0,0,0,4,4,0,1;
  1,0,2,0,0,0,0,0,0,1;
  1,0,0,0,0,0,5,0,0,1;
  1,0,0,0,0,0,0,0,0,1;
  1,1,1,1,1,1,1,1,1,1; ];

grid( opg(2), opg(1) ) = 0;

p = zeros( 2, 1 );

% Maze window.
h_3d = figure( 1 );
clf
pos = get( h_3d, 'position' );
set( h_3d,...
    'backingstore', 'off',...
    'menubar', 'none',...
    'position', [pos(1), pos(2), fov.s(1), fov.s(2)],...
    'keypressfcn', 'key_event',...
    'tag', 'maze_window' );
h_ax_3d = gca;
set( h_ax_3d, 'position', [0, 0, 1, 1], 'clipping', 'off' );
view(3)
axis equal
axis off
% $$$ hold on

tx = zeros(tex_h,gcs(1),3);
k = 1;
for i = 1:ng(1)
  for j = 1:ng(2)
    n = grid(j,i);
    if (n)
      x = [(i - 1) * gcs(2), (i - 1) * gcs(2);...
	(i - 1) * gcs(2), (i - 1) * gcs(2)];
      y = [(j - 0) * gcs(1), (j - 1) * gcs(1);...
	(j - 0) * gcs(1), (j - 1) * gcs(1)];
      z = [0, 0; tex_h, tex_h];
      tx(:,:,:) = texture(n,:,:,:);
      h_surf(k) = surface(x, y, z, 'facecolor', 'texture',...
	  'cdata', tx, 'edgecolor', 'none' );
      k = k + 1;
      x = [(i - 1) * gcs(2), (i - 0) * gcs(2);...
	(i - 1) * gcs(2), (i - 0) * gcs(2)];
      y = [(j - 1) * gcs(1), (j - 1) * gcs(1);...
	(j - 1) * gcs(1), (j - 1) * gcs(1)];
      z = [0, 0; tex_h, tex_h];
      tx(:,:,:) = texture(n,:,:,:);
      h_surf(k) = surface(x, y, z, 'facecolor', 'texture',...
	  'cdata', tx, 'edgecolor', 'none' );
      k = k + 1;
      x = [(i - 0) * gcs(2), (i - 0) * gcs(2);...
	(i - 0) * gcs(2), (i - 0) * gcs(2)];
      y = [(j - 0) * gcs(1), (j - 1) * gcs(1);...
	(j - 0) * gcs(1), (j - 1) * gcs(1)];
      z = [0, 0; tex_h, tex_h];
      tx(:,:,:) = texture(n,:,:,:);
      h_surf(k) = surface(x, y, z, 'facecolor', 'texture',...
	  'cdata', tx, 'edgecolor', 'none' );
      k = k + 1;
      x = [(i - 1) * gcs(2), (i - 0) * gcs(2);...
	(i - 1) * gcs(2), (i - 0) * gcs(2)];
      y = [(j - 0) * gcs(1), (j - 0) * gcs(1);...
	(j - 0) * gcs(1), (j - 0) * gcs(1)];
      z = [0, 0; tex_h, tex_h];
      tx(:,:,:) = texture(n,:,:,:);
      h_surf(k) = surface(x, y, z, 'facecolor', 'texture',...
	  'cdata', tx, 'edgecolor', 'none' );
      k = k + 1;
    end
  end
end

g_quit = 0;
g_is_map_fig = 0;
g_event = '';
while (~g_quit)
  t0 = clock;
  oaa = oa * pi / 180;
  ov = [cos(oaa), -sin(oaa); sin(oaa), cos(oaa)] * [1; 0];
  set( h_ax_3d,...
      'projection', 'perspective',...
      'cameraposition', [op(:)', tex_h/2],...
      'cameratarget', [op(:)' + ov(:)', tex_h/2],...
      'cameraupvector', [0 0 1],...
      'cameraviewangle', fov.a );
  % Show maze.
  drawnow
  % Process events.
  set( h_3d, 'keypressfcn', '' );
  ll = length( g_event );
  if (ll)
    fps = 1 / etime( clock, t0 )
  end
  i = 1;
  no_events = ~ll;
  while (1)
    if (no_events)
      break
    end
    ch = g_event(i);
    % Remove autorepeat.
    while (g_event(i) == ch)
      i = i + 1;
      if (i > ll)
	no_events = 1;
	break
      end
    end
    switch (ch)
      case '8', % Move forward.
	op0 = op;
	op = op + mov.forward * ov;
	opg = floor( op ./ gcs ) + 1;
	if (any( [opg(1) <= 0, opg(2) <= 0,...
		opg(1) > ng(1), opg(2) > ng(2)] ) |...
	      (grid( opg(2), opg(1) )))
	  op = op0;
	end
      case '5', % Move backward.
	op0 = op;
	op = op + mov.backward * ov;
	opg = floor( op ./ gcs ) + 1;
	if (any( [opg(1) <= 0, opg(2) <= 0,...
		opg(1) > ng(1), opg(2) > ng(2)] ) |...
	      (grid( opg(2), opg(1) )))
	  op = op0;
	end
      case '4', % Strafe left.
	op0 = op;
	op = op + mov.strafe * ov;
	opg = floor( op ./ gcs ) + 1;
	if (any( [opg(1) <= 0, opg(2) <= 0,...
		opg(1) > ng(1), opg(2) > ng(2)] ) |...
	      (grid( opg(2), opg(1) )))
	  op = op0;
	end
      case '6', % Strafe right.
	op0 = op;
	op = op + mov.strafe * ov;
	opg = floor( op ./ gcs ) + 1;
	if (any( [opg(1) <= 0, opg(2) <= 0,...
		opg(1) > ng(1), opg(2) > ng(2)] ) |...
	      (grid( opg(2), opg(1) )))
	  op = op0;
	end
      case '7', % Turn left.
	oa = oa + mov.turn;
      case '9', % Turn right.
	oa = oa - mov.turn;
      case 'i', % Info.
	disp( ['Position: ' num2str( op(:)' )] );
	disp( ['Grid position: ' num2str( opg(:)' )] );
	disp( ['View angle: ' num2str( oa )] );
	disp( ['FPS: ' num2str( fps )] );
	disp( ' ' )
      case 'm', % Map on/off.
	h_map = figure( 2 );
	g_is_map_fig = ~g_is_map_fig;
	if (g_is_map_fig)
	  plot_map( h_map, op, oa, grid, gcs, ng, fov );
	else
	  close( h_map );
	end
      case 'r', % Refresh map.
	if (~g_is_map_fig)
	  h_map = figure( 2 );
	  g_is_map_fig = ~g_is_map_fig;
	end
	plot_map( h_map, op, oa, grid, gcs, ng, fov );
      case 's', % Set observer position on map.
	if (~g_is_map_fig)
	  h_map = figure( 2 );
	  g_is_map_fig = ~g_is_map_fig;
	end
	plot_map( h_map, op, oa, grid, gcs, ng, fov );
	[p(1), p(2), b] = ginput( 1 );
	pg = floor( p ./ gcs ) + 1;
	if (~(any( [pg(1) <= 0, pg(2) <= 0,...
		pg(1) > ng(1), pg(2) > ng(2)] )) & (b == 1))
	  if  (~grid( pg(2), pg(1) ))
	    op = p;
	  end
	elseif (b ~= 1)
	  oa = (1-2*(p(2) < op(2)))...
	      * 180 / pi * acos( (p(1) - op(1)) / norm( p - op ) );
	end
      case 'q', % Quit.
	g_quit = 1;
    end
  end
  g_event = '';
  set( h_3d, 'keypressfcn', 'key_event' );
end
