% function [ni] = render_maze( op, oa, grid, gcs, ng, fov )
% input:
%   op .. observer point
%   oa .. observer view vector angle
%   grid .. maze grid
%   gcs .. grid cell sizes
%   ng .. grid sizes
%   smap .. scale map
%   texture .. texture field
% output:
%   ni .. number of intersections (== rendered columns)
% input & output:
%   fov .. field of view information, contains rendered image
function [ni, fov] = render_maze( op, oa, grid, gcs, ng, smap, texture, fov )
global g_angle

  ni = 0;
  col = zeros( smap.lmax-smap.lmin+1, 3 );
  % Init maze.
  fov.img(1:round(fov.s(2)/2),:,:) = 1;
  fov.img(round(fov.s(2)/2):fov.s(2),:,:) = 0;
  
  for i = 1:fov.s(1)
    oaa = oa + g_angle(i);
    [ok, li, cp, tc] = trace_ray( oaa, op, grid, gcs, ng );
    % Render i-th maze column.
    if (ok)
      ni = ni + 1;
      tex = grid( cp(2), cp(1) );
      ll = norm( li - op );
      l = round( fov.sd * gcs(2) / (ll)) + 1;
      if (l > smap.lmax)
	l = smap.lmax;
      end
      aux = round((fov.s(2) - l)/2);
      map_row = l-smap.lmin+1;
      if (l > fov.s(2))
	cbot = 1 - aux;
	ctop = cbot + fov.s(2) - 1;
	col(cbot:ctop,:) = scale_line_rgb( squeeze( texture(tex,:,tc,:) ),...
	    smap.map(map_row,cbot:ctop) );
	fov.img(1:fov.s(2),i,:) = col(cbot:ctop,:);
      else
	cbot = 1 + aux;
	ctop = cbot + l - 1;
	col(1:l,:) = scale_line_rgb( squeeze( texture(tex,:,tc,:) ),...
	    smap.map(map_row,1:l) );
	fov.img(cbot:ctop,i,:) = col(1:l,:);
      end
    end
  end
