% function [ok, li, cp, tc] = trace_ray( angle, op, grid, gcs, ng )
% input:
%   angle .. obsetver view vector angle
%   op .. observer point
%   grid .. maze grid
%   gcs .. grid cell sizes
%   ng .. grid sizes
% output:
%   ok .. something was hit?
%   li .. line intersection point
%   cp .. hit cell
%   tc .. texture column corresponding to li
function [ok, li, cp, tc] = trace_ray( angle, op, grid, gcs, ng )
global g_zero

% Observer angle, vector.
oa = angle * pi / 180;
ov = [cos(oa), -sin(oa); sin(oa), cos(oa)] * [1; 0];

% Observer point in grid coordinates.
opg = floor( op ./ gcs );

% Check point vector.
cpv = [-(ov(1) < -g_zero); -(ov(2) < -g_zero)];
cpv = repmat( cpv, 1, max( ng ) );

gcs = repmat( gcs, 1, max( ng ) );

% X-line intersections.
ok_x = 0;
if (ov(1) > g_zero)
  i0 = opg(1) + 1;
  di = 1;
  i1 = ng(1);
elseif (ov(1) < -g_zero)
  i0 = opg(1);
  di = -1;
  i1 = 0;
else
  di = 0;
end
if (di ~= 0)
  iv = i0:di:i1;
  liv = length(iv);
  s = (iv * gcs(1,1) - op(1)) / ov(1);
  xl(1,:) = round( op(1) + s * ov(1) );
  xl(2,:) = round( op(2) + s * ov(2) );
  % Checkpoint.
  xc = floor( (xl + cpv(:,1:liv)) ./ gcs(:,1:liv) ) + 1;
  for i = 1:liv
    if (any( [xc(1,i) <= 0, xc(2,i) <= 0,...
	    xc(1,i) > ng(1), xc(2,i) > ng(2)] ))
      break;
    end
    cell = grid(xc(2,i),xc(1,i));
    if (cell ~= 0)
      ok_x = 1;
      xli = xl(:,i);
      xcp = xc(:,i);
      break
    end
  end
end

% Y-line intersections.
ok_y = 0;
if (ov(2) > g_zero)
  i0 = opg(2) + 1;
  di = 1;
  i1 = ng(2);
elseif (ov(2) < -g_zero)
  i0 = opg(2);
  di = -1;
  i1 = 0;
else
  di = 0;
end
if (di ~= 0)
  iv = i0:di:i1;
  liv = length(iv);
  s = (iv * gcs(2,1) - op(2)) / ov(2);
  yl(1,:) = round( op(1) + s * ov(1) );
  yl(2,:) = round( op(2) + s * ov(2) );
  % Checkpoint.
  yc = floor( (yl + cpv(:,1:liv)) ./ gcs(:,1:liv) ) + 1;
  for i = 1:liv
    if (any( [yc(1,i) <= 0, yc(2,i) <= 0,...
	    yc(1,i) > ng(1), yc(2,i) > ng(2)] ))
      break;
    end
    cell = grid(yc(2,i),yc(1,i));
    if (cell ~= 0)
      ok_y = 1;
      yli = yl(:,i);
      ycp = yc(:,i);
      break
    end
  end
end

ok = 1;
tor = 0;
if (ok_x & ok_y)
  xdist = norm( xli - op );
  ydist = norm( yli - op );
  if (xdist < ydist)
    li = xli;
    cp = xcp;
    tc = li(2) - (cp(2) - 1) * gcs(2);
    tor = (ov(1) > g_zero);
  else
    li = yli;
    cp = ycp;
    tc = li(1) - (cp(1) - 1) * gcs(1);
    tor = (ov(2) < -g_zero);
  end
elseif (ok_x)
  li = xli;
  cp = xcp;
  tc = li(2) - (cp(2) - 1) * gcs(2);
  tor = (ov(1) > g_zero);
elseif (ok_y)
  li = yli;
  cp = ycp;
  tc = li(1) - (cp(1) - 1) * gcs(1);
  tor = (ov(2) < -g_zero);
else
  ok = 0;
  li = [0; 0];
  cp = [0; 0];
  tc = 0;
end
% Fix zero tc.
tc = (ok & ~tc) + tc;
% Fix texture orientation.
tc = tor * (gcs(1) - tc + 1) + (1 - tor) * tc;
